//
// svg image display demo program for the Fast Light Tool Kit (FLTK).
//
// Copyright (c) 2016 by Albrecht Schlosser
//
// This library is free software. Distribution and use rights are outlined in
// the file "COPYING" which should have been included with this file.  If this
// file is missing or damaged, see the license at:
//
//     http://www.fltk.org/COPYING.php
//
// Please report all bugs and problems on the following page:
//
//     http://www.fltk.org/str.php

//
// Build instructions:
//
// Download the two nanosvg header files nanosvg.h and nanosvgrast.h
// (see links below) to the same folder as this file (svg.cxx).
//
// Compile with: fltk-config --use-images --compile svg.cxx
//
// Usage: ./svg
//
// Full implementation of class Fl_SVG_Image, as derived from Fl_RGB_Image
// Demo shows integration with FLTK
//
//

#include <FL/Fl.H>
#include <FL/Fl_Double_Window.H>
#include <FL/Fl_Box.H>
#include <FL/Fl_RGB_Image.H>
#include <FL/Fl_Shared_Image.H>
#include <FL/Fl_File_Chooser.H>
#include <errno.h>

class Fl_SVG_Box;			// displays scaled SVG image

/*
  The entire nanosvg library code is included by defining the two macros
  NANOSVG*_IMPLEMENTATION before including the headers.

  Source: https://github.com/memononen/nanosvg.git

  Direct download links:

  https://raw.githubusercontent.com/memononen/nanosvg/master/src/nanosvg.h
  https://raw.githubusercontent.com/memononen/nanosvg/master/src/nanosvgrast.h
*/

#include <stdio.h>
#include <string.h>
#include <math.h>

#define NANOSVG_ALL_COLOR_KEYWORDS	// Include full list of color keywords.
#define NANOSVG_IMPLEMENTATION		// Expands implementation
#include "nanosvg.h"

#define NANOSVGRAST_IMPLEMENTATION	// Expands implementation
#include "nanosvgrast.h"


class Fl_SVG_Image : public Fl_RGB_Image {
private:
  char *svg_data;
  NSVGimage* svg_image;
  NSVGrasterizer* rast;
  bool to_desaturate;
  Fl_Color average_color;
  float average_weight;
  float svg_scaling(int W, int H) {
    float f1 = float(W) / int(svg_image->width+0.5);
    float f2 = float(H) / int(svg_image->height+0.5);
    return (f1 < f2) ? f1 : f2;
  }
  void rasterize() {
    float f = svg_scaling(w(), h());
    array = new unsigned char[w()*h()*4];
    nsvgRasterize(rast, svg_image, 0, 0, f, (uchar* )array, w(), h(), w()*4);
    alloc_array = 1;
    data((const char * const *)&array, 1);
    d(4);
    if (to_desaturate) Fl_RGB_Image::desaturate();
    if (average_weight < 1) Fl_RGB_Image::color_average(average_color, average_weight);
  }

public:
  Fl_SVG_Image(const char *filename, const char *buffer = 0) : Fl_RGB_Image(NULL, 0, 0, 4) {
    svg_data = NULL;
    svg_image = NULL;
    to_desaturate = false;
    average_weight = 1;
    size_t size;
    char *filedata = NULL;
    if (filename) {
      FILE *fp = fopen(filename, "rb");
      if (fp) {
        fseek(fp, 0, SEEK_END);
        size = ftell(fp);
        fseek(fp, 0, SEEK_SET);
        filedata = (char*)malloc(size+1);
        if (filedata) {
          if (fread(filedata, 1, size, fp) == size) filedata[size] = '\0';
          else {
            free(filedata);
            filedata = NULL;
          }
        }
        fclose(fp);
      } else ld(ERR_FILE_ACCESS);
    } else {
      filedata = strdup(buffer);
    }
    if (filedata) {
      svg_data = strdup(filedata);
      svg_image = nsvgParse(filedata, "px", 96);
      free(filedata);
      if (svg_image->width == 0 || svg_image->height == 0) {
        ld(ERR_FORMAT);
      }
      rast = nsvgCreateRasterizer();
      w(svg_image->width + 0.5);
      h(svg_image->height + 0.5);
    }
  }
  virtual ~Fl_SVG_Image() {
    char *p = (char*)svg_data; if (p) free(p);
    nsvgDelete(svg_image);
    nsvgDeleteRasterizer(rast);
  }
  virtual void draw(int X, int Y, int W, int H, int cx=0, int cy=0) {
    if (!array) rasterize();
    Fl_RGB_Image::draw(X,Y,W,H,cx,cy);
  }
  virtual Fl_Image *copy(int W, int H) {
    Fl_SVG_Image *svg2 = new Fl_SVG_Image(NULL, svg_data);
    float f = (W-1)/float(H+1);
    float F = float(W+1)/(H-1);
    float m = float(svg2->w()) / svg2->h();
    if ( f <= m && m <= F) {
      svg2->w(W);
      svg2->h(H);
    } else {
      svg2->resize(W, H);
    }
    return svg2;
  }
  void resize(int W, int H) {
    if (array) {
      delete[] array;
      array = NULL;
    }
    uncache();
    float f = svg_scaling(W, H);
    int w1 = int( int(svg_image->width+0.5)*f +0.5 );
    int h1 = int( int(svg_image->height+0.5)*f +0.5 );
    w(w1); h(h1);
  }
  virtual void desaturate() {
    to_desaturate = true;
    Fl_RGB_Image::desaturate();
  }
  virtual void color_average(Fl_Color c, float i) {
    average_color = c;
    average_weight = i;
    Fl_RGB_Image::color_average(c, i);
  }
};

// To be ultimately added to fl_register_images()
Fl_Image *svg_handler(const char *name, uchar *header, int headerlen) {
  if ( (headerlen > 5 && memcmp(header, "<?xml", 5) == 0) ||
      memcmp(header, "<svg", 4) == 0)
    return new Fl_SVG_Image(name);
  return NULL;
}


class image_box : public Fl_Box { // a box with a resizing image inside
public:
  image_box(int x, int y, int w , int h, Fl_Shared_Image *shared) : Fl_Box(x,y,w,h) {
    shared->scale(w, h);
    image(shared);
    align(FL_ALIGN_CENTER|FL_ALIGN_INSIDE);
    Fl_Color bg = fl_rgb_color(0xff, 0xdd, 0xff);		// bright pink bg color
    box(FL_FLAT_BOX);
    color(bg);
  }
  virtual void resize(int x, int y, int w, int h) {
    Fl_Shared_Image *shared = (Fl_Shared_Image*)image();
    if (shared) {
      shared->scale(w, h);
    }
    Fl_Box::resize(x,y,w,h);
  }
};


int main(int argc, char **argv) {
  fl_register_images();
  Fl_Shared_Image::add_handler(svg_handler);
  while (1) {
    const char *filename = fl_file_chooser("Select an image file","*.{svg,jpg,png,pxm}", NULL);
    if (!filename) break;
    Fl_Shared_Image *shared = Fl_Shared_Image::get(filename);
    if (!shared) continue;
    //shared->desaturate();
    //shared->color_average(FL_GREEN, .5);
    
    Fl_Double_Window *window = new Fl_Double_Window(620,620);
    window->color(FL_WHITE);
    image_box *box = new image_box(5, 5, window->w()-10, window->h()-10, shared);
    window->end();
    window->resizable(box);
    window->show(argc , argv);
    Fl::run();
    shared->release();
    delete window;
  }
  return 0;
}
